/*
	SOS Tester - [S]imple [O]S [S]cheduler Tester
	(c) Mikhail Panasyuk aka BeAR, 2009. E-mail: otoko@yandex.ru
	
	Now you can tell which one (Haiku, BeOS, Zeta...) is fair enough ;)
	Run this program on different OSes and compare results.
	
	Original concept and results' interpretation:
	
		Fair scheduler should fairly redistribute CPU time between time-sharing
		threads. Greater priority means more CPU time. And no threads should be
		left behind even if they were much lower priority than others.
		
		This program simulates situations when there are lots of threads
		running at different priorities and all of them do same simple task -
		infinitely increment their individual counters by 1 starting from 0.
		
		After some time program shuts all working threads down and prints out
		values each counter ended with along with priority it was incremented
		at.
		
		Now you have data and can judge if CPU time was distributed fair enough!
	
	
	Take note that working thread priorities should be somewhere in between of
	most system / app threads (usually up to 20-22) and real-time threads
	(100-120).
	
	Main thread should have real-time priority because this way there is some
	confidence that working threads will not run out of control.
	
	Use	at your own risk, this program may cause system crash or even burn your
	CPU(s), no warranty is provided whatsoever.
	
	Compile like this: gcc sos_tester.cpp -lbe -lroot -o sos_tester
	Run like this: sos_tester -t 60000 -n 60 -p 30 99 >sos_tester_output.txt
		or simply: sos_tester
	
	Have fun! ;)
*/


#include <OS.h>
#include <String.h>

#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>


/* Defaults section -- Begin */

// length of lifecycle
// milliseconds
#define DEATH_TIMER_INT32 10000 // 10 seconds
#define MIN_DEATH_TIMER_INT32 1000 // 1 second
#define MAX_DEATH_TIMER_INT32 3600000 // 1 hour

// threads' number regulations
#define NUM_THREADS 20
#define MIN_NUM_THREADS 2
#define MAX_NUM_THREADS 4096

// threads' priority regulations
#define L_PRIORITY 30
#define MIN_L_PRIORITY 1
#define MAX_L_PRIORITY 119

#define H_PRIORITY 99
#define MIN_H_PRIORITY 1
#define MAX_H_PRIORITY 119

#define PRIORITY_FORMULA(i,n,low,high) (int)(low*1.0+((high-low)*1.0/(n-1))*i)

/* Defaults section -- End */


#define IS_IN_RANGE(i,min,max) (i >= min && i <= max)

typedef struct {
	thread_id	id;
	int			priority;
	int64		counter;
} sos_thread_info;

sos_thread_info threads[MAX_NUM_THREADS];

// Options

bigtime_t deathTimer;
int lowestPriority;
int highestPriority;
int threadsNumber;
int threadsSpawned;

bigtime_t delta, minWorktime;

// Print functions

void print_usage(char* appName);
void print_invalid_args(char* appName);
void print_settings();
void print_real();
void print_results_table();

void get_default_settings();
bool get_user_settings(int argc, char** argv);

void crunch_func(int64* counter)
{
	while (1)
		*counter += 1;
}

int main(int argc, char** argv)
{
	get_default_settings();
	if (!get_user_settings(argc, argv)) {
		return 1;
	}
	print_settings();

	threadsSpawned = 0;
	for (int i=0; i<threadsNumber; i++) {
		BString thread_name("crunch_thread ");
		thread_name << i;
		threads[i].priority = PRIORITY_FORMULA(
			i, threadsNumber, lowestPriority, highestPriority
		);
		
		threads[i].id = spawn_thread((thread_func)crunch_func,
			thread_name.String(), threads[i].priority, &(threads[i].counter));
		
		if ((threads[i].id) < B_OK) {
			// handle error
			fprintf(stderr, "Warning: can't spawn more than %i threads.\n", i);
			break;
		} else {
			threads[i].counter = 0;
			threadsSpawned++;
		}
	}

	set_thread_priority(find_thread(NULL), B_REAL_TIME_PRIORITY);
		// tell system who is the boss ;)

	bigtime_t event0, event1, event2, event3;

	event0 = system_time();

	for (int i=threadsSpawned-1; i>=0; i--) {
		resume_thread(threads[i].id);
	}

	event1 = system_time();

	snooze(deathTimer);

	event2 = system_time();

	for (int i=0; i<threadsSpawned; i++) {
		kill_thread(threads[i].id);
	}

	event3 = system_time();

	delta = (event1 - event0) + (event3 - event2);
	minWorktime = event2 - event1;
	
	print_real();
	printf("\n");
	print_results_table();

	return 0;
}

void print_settings()
{
	printf("Settings: work time %Lu microseconds, %i working threads.\n",
		deathTimer, threadsNumber);
}

void print_real()
{
	printf("Real work time %Lu microseconds, %i working threads.\n",
		minWorktime, threadsSpawned);
	printf("Inaccuracy if SMP and more than one CPU enabled: %f%%.\n",
		(delta*1.0 / minWorktime*1.0) * 100);
}

void print_results_table()
{
	int64 sum = 0;
	for (int i=0; i<threadsSpawned; i++) {
		sum += threads[i].counter;
	}
	printf("  Number  Priority            Value    Share (%%)\n"
		   "--------------------------------------------------\n");
	for (int i=0; i<threadsSpawned; i++) {
		double percent = (threads[i].counter*1.0) / (sum*1.0) * 100;
		printf("%8i %9i %16Lu %12f\n",
			i, threads[i].priority, threads[i].counter, percent);
	}
}

void print_invalid_args(char* appName)
{
	fprintf(stderr, "Invalid arguments. Type '%s -h' to get help.\n", appName);
}

void print_usage(char* appName)
{
	fprintf(stderr,
		"Usage: %s [-h] [-t <time>] [-n <number>] [-p <low> <high>]\n\n"
		"Defaults: -t %i -n %i -p %i %i\n\n"
		"General:\n"
		"  -h, --help\n"
		"    : show this help\n\n"
		"Test case:\n"
		"  -t, --timer <time>\n"
		"    : set work time in milliseconds,\n"
		"      <time> - between %i and %i\n\n"
		"  -n, --threads <number>\n"
		"    : set number of threads,\n"
		"      <number> - between %i and %i\n\n"
		"  -p, --priority <low> <high>\n"
		"    : set range of threads' priorities,\n"
		"      <low> - between %i and %i\n"
		"      <high> - between %i and %i\n\n",
		appName,
		DEATH_TIMER_INT32, NUM_THREADS, L_PRIORITY, H_PRIORITY,
		MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32,
		MIN_NUM_THREADS, MAX_NUM_THREADS,
		MIN_L_PRIORITY, MAX_L_PRIORITY,
		MIN_H_PRIORITY, MAX_H_PRIORITY
	);
}

void get_default_settings()
{
	deathTimer = DEATH_TIMER_INT32 * 1000; // conversion to microseconds
	threadsNumber = NUM_THREADS;
	lowestPriority = L_PRIORITY;
	highestPriority = H_PRIORITY;
}

bool get_user_settings(int argc, char** argv)
{
	int optc;
	static struct option long_options[] = {
		{"help",		no_argument,		0,	'h'},
		{"timer",		required_argument,	0,	't'},
		{"threads",		required_argument,	0,	'n'},
		{"priority",	required_argument,	0,	'p'},
		{0, 0, 0, 0}
	};

	// TODO: filer wrong --* arguments
	
	while ((optc = getopt_long (argc, argv, "h?tnp", long_options,
		NULL)) != EOF)
	{
		switch (optc) {
			case 'h':
			case '?':
				print_usage(argv[0]);
				return false;
			case 't':
				if( !optarg && optind < argc) {
					int _deathTimer = atoi(argv[optind]);
					if (IS_IN_RANGE(_deathTimer,
							MIN_DEATH_TIMER_INT32, MAX_DEATH_TIMER_INT32))
					{
						deathTimer = _deathTimer * 1000;
							// conversion to microseconds
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'n':
				if( !optarg && optind < argc) {
					int _threadsNumber = atoi(argv[optind]);
					if (IS_IN_RANGE(_threadsNumber,
							MIN_NUM_THREADS, MAX_NUM_THREADS))
					{
						threadsNumber = _threadsNumber;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			case 'p':
				if( !optarg && optind < argc - 1) {
					int _lowestPriority = atoi(argv[optind]);
					int _highestPriority = atoi(argv[optind + 1]);
					if (IS_IN_RANGE(_lowestPriority,
							MIN_L_PRIORITY, MAX_L_PRIORITY) &&
						IS_IN_RANGE(_highestPriority,
							MIN_H_PRIORITY, MAX_H_PRIORITY) &&
						_lowestPriority <= _highestPriority)
					{
						lowestPriority = _lowestPriority;
						highestPriority = _highestPriority;
					} else {
						print_invalid_args(argv[0]);
						return false;
					}
				} else {
					print_invalid_args(argv[0]);
					return false;
				}
				break;
			default:
				break;
		}
	}

	return true;
}
